@@ -1,13 +1,14 @@
-/* $Id: dom.c,v 1.44 2002/09/12 18:16:15 phish Exp $ */
+/* $Id: dom.c,v 1.51 2003/02/04 23:01:32 phish Exp $ */
 #include <libxml/tree.h>
 #include <libxml/encoding.h>
 #include <libxml/xmlerror.h>
 #include <libxml/xmlmemory.h>
 #include <libxml/parser.h>
+#include <libxml/parserInternals.h>
 #include <libxml/xmlIO.h>
 #include <libxml/xpath.h>
 #include <libxml/xpathInternals.h>
-
+#include <libxml/globals.h>
 #include <stdio.h>
 
 #define warn(string) fprintf(stderr, string) 
@@ -18,6 +19,94 @@
 #define xs_warn(string)
 #endif
 
+/**
+ * NAME domParseChar
+ * TYPE function
+ * SYNOPSIS
+ *   int utf8char = domParseChar( curchar, &len );
+ *
+ * The current char value, if using UTF-8 this may actually span
+ * multiple bytes in the given string. This function parses an utf8
+ * character from a string into a UTF8 character (an integer). It uses
+ * a slightly modified version of libxml2's character parser. libxml2
+ * itself does not provide any function to parse characters dircetly
+ * from a string and test if they are valid utf8 characters.
+ *
+ * XML::LibXML uses this function rather than perls native UTF8
+ * support for two reasons:
+ * 1) perls UTF8 handling functions often lead to encoding errors,
+ *    which partly comes, that they are badly documented.
+ * 2) not all perl versions XML::LibXML intends to run with have native
+ *    UTF8 support.
+ *
+ * domParseChar() allows to use the very same code with all versions
+ * of perl :)
+ *
+ * Returns the current char value and its length
+ *
+ * NOTE: If the character passed to this function is not a UTF
+ * character, the return value will be 0 and the length of the
+ * character is -1!
+ */
+int
+domParseChar( xmlChar *cur, int *len ) 
+{
+    unsigned char c;
+	unsigned int val;
+
+	/*
+	 * We are supposed to handle UTF8, check it's valid
+	 * From rfc2044: encoding of the Unicode values on UTF-8:
+	 *
+	 * UCS-4 range (hex.)           UTF-8 octet sequence (binary)
+	 * 0000 0000-0000 007F   0xxxxxxx
+	 * 0000 0080-0000 07FF   110xxxxx 10xxxxxx
+	 * 0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx 
+	 *
+	 * Check for the 0x110000 limit too
+	 */
+    
+    if ( cur == NULL || *cur == 0 ) {
+        *len = 0;
+        return(0);
+    }
+    
+    c = *cur;
+    if ( c & 0x80 ) { 
+        if ((c & 0xe0) == 0xe0) {
+            if ((c & 0xf0) == 0xf0) {
+                /* 4-byte code */
+                *len = 4;
+                val = (cur[0] & 0x7) << 18;
+                val |= (cur[1] & 0x3f) << 12;
+                val |= (cur[2] & 0x3f) << 6;
+                val |= cur[3] & 0x3f;
+            } else {
+                /* 3-byte code */
+                *len = 3;
+                val = (cur[0] & 0xf) << 12;
+                val |= (cur[1] & 0x3f) << 6;
+                val |= cur[2] & 0x3f;
+            }
+	    } else {
+            /* 2-byte code */
+            *len = 2;
+            val = (cur[0] & 0x1f) << 6;
+            val |= cur[1] & 0x3f;
+	    }
+        if ( !IS_CHAR(val) ) {
+            *len = -1;
+            return(0);
+        }
+	    return(val);
+    }
+    else {
+        /* 1-byte code */
+	    *len = 1;
+        return((int)c); 
+    }
+}
+
 /**
  * Name: domReadWellBalancedString
  * Synopsis: xmlNodePtr domReadWellBalancedString( xmlDocPtr doc, xmlChar *string )
@@ -267,11 +356,18 @@ domImportNode( xmlDocPtr doc, xmlNodePtr node, int move ) {
         }
     }
 
+
     /* tell all children about the new boss */ 
     if ( node && doc && node->doc != doc ) {
         xmlSetTreeDoc(return_node, doc);
     }
- 
+
+    if ( doc != NULL 
+         && return_node != NULL
+         && return_node->type != XML_ENTITY_REF_NODE ) {
+        xmlReconciliateNs(doc, return_node);     
+    }
+
     return return_node;
 }
 
@@ -285,17 +381,67 @@ domImportNode( xmlDocPtr doc, xmlNodePtr node, int move ) {
  **/
 xmlChar*
 domName(xmlNodePtr node) {
-    xmlChar *qname = NULL; 
-    if ( node ) {
-        if (node->ns != NULL && node->ns->prefix != NULL) {
-            xmlChar *tname = xmlStrdup( node->ns->prefix );
-            tname = xmlStrcat( tname , ":" );
-            tname = xmlStrcat( tname , node->name );
-            qname = tname;
-        } 
-        else {
-            qname = xmlStrdup( node->name );
+    const xmlChar *prefix = NULL;
+    const xmlChar *name   = NULL;
+    xmlChar *qname        = NULL; 
+    
+    if ( node == NULL ) {
+        return NULL;
+    }
+
+    switch ( node->type ) {
+    case XML_TEXT_NODE :
+    case XML_COMMENT_NODE :
+    case XML_CDATA_SECTION_NODE :
+    case XML_XINCLUDE_START :
+    case XML_XINCLUDE_END :
+    case XML_ENTITY_REF_NODE :
+    case XML_ENTITY_NODE :
+    case XML_DTD_NODE :
+    case XML_ENTITY_DECL :
+    case XML_DOCUMENT_TYPE_NODE :
+    case XML_PI_NODE :
+    case XML_NOTATION_NODE :
+    case XML_NAMESPACE_DECL :
+        name = node->name;
+        break;
+
+    case XML_DOCUMENT_NODE :
+    case XML_HTML_DOCUMENT_NODE :
+    case XML_DOCB_DOCUMENT_NODE :
+        name = "document";
+        break;
+
+    case XML_DOCUMENT_FRAG_NODE :
+        name = "document_fragment";
+        break;
+
+    case XML_ELEMENT_NODE :
+    case XML_ATTRIBUTE_NODE :
+        if ( node->ns != NULL ) {
+            prefix = node->ns->prefix;
         }
+        name = node->name;
+        break;
+
+    case XML_ELEMENT_DECL :
+        prefix = ((xmlElementPtr) node)->prefix;
+        name = node->name;
+        break;
+
+    case XML_ATTRIBUTE_DECL :
+        prefix = ((xmlAttributePtr) node)->prefix;
+        name = node->name;
+        break;
+    }
+
+    if ( prefix != NULL ) {
+        qname = xmlStrdup( prefix );
+        qname = xmlStrcat( qname , (const xmlChar *) ":" );
+        qname = xmlStrcat( qname , name );
+    } 
+    else {
+        qname = xmlStrdup( node->name );
     }
 
     return qname;
@@ -359,7 +505,9 @@ domAppendChild( xmlNodePtr self,
         newChild->parent= self;
     }
  
-    xmlReconciliateNs(self->doc, newChild);     
+    if ( newChild->type != XML_ENTITY_REF_NODE ) {
+        xmlReconciliateNs(self->doc, newChild);     
+    }
 
     return newChild;
 }
@@ -413,6 +561,11 @@ domReplaceChild( xmlNodePtr self, xmlNodePtr new, xmlNodePtr old ) {
         domRemoveChild( self, old );
         domAppendChild( self, new );
     }
+    else if ( new->type == XML_DOCUMENT_FRAG_NODE 
+              && new->children == NULL ) {
+        /* want to replace with an empty fragment, then remove ... */
+        domRemoveChild( self, old );
+    }
     else {
         domAddNodeToList(new, old->prev, old->next );
         old->parent = old->next = old->prev = NULL;    
@@ -459,7 +612,9 @@ domInsertBefore( xmlNodePtr self,
     }
     
     domAddNodeToList(newChild, refChild->prev, refChild);
-     xmlReconciliateNs(self->doc, newChild);     
+    if ( newChild->type != XML_ENTITY_REF_NODE ) {
+        xmlReconciliateNs(self->doc, newChild);     
+    }
 
     return newChild;
 }
@@ -502,7 +657,9 @@ domInsertAfter( xmlNodePtr self,
     }
 
     domAddNodeToList(newChild, refChild, refChild->next);
-     xmlReconciliateNs(self->doc, newChild);     
+    if ( newChild->type != XML_ENTITY_REF_NODE ) {
+        xmlReconciliateNs(self->doc, newChild);     
+    }
 
     return newChild;
 }
@@ -548,7 +705,9 @@ domReplaceNode( xmlNodePtr oldNode, xmlNodePtr newNode ) {
         domAddNodeToList( newNode, prev,  next );
     }
 
-     xmlReconciliateNs(newNode->doc, newNode); 
+    if ( newNode->type != XML_ENTITY_REF_NODE ) {
+        xmlReconciliateNs(newNode->doc, newNode); 
+    }
 
     return oldNode;
 }
@@ -610,7 +769,7 @@ domSetNodeValue( xmlNodePtr n , xmlChar* val ){
     if ( n == NULL ) 
         return;
     if ( val == NULL ){
-        val = "";
+        val = (xmlChar *) "";
     }
   
     if( n->type == XML_ATTRIBUTE_NODE ){
@@ -879,3 +1038,4 @@ domNodeNormalizeList( xmlNodePtr nodelist )
     }
     return(1);
 }
+
