@@ -1,4 +1,4 @@
-# $Id: Builder.pm,v 1.22 2002/06/25 18:16:54 phish Exp $
+# $Id: Builder.pm,v 1.28 2003/02/04 23:01:33 phish Exp $
 
 package XML::LibXML::SAX::Builder;
 
@@ -10,6 +10,23 @@ sub new {
     return bless {@_}, $class;
 }
 
+sub result { $_[0]->{LAST_DOM}; }
+
+sub done {
+    my ($self) = @_;
+    my $dom = $self->{DOM};
+    $dom = $self->{Parent} unless defined $dom; # this is for parsing document chunks
+
+    delete $self->{NamespaceStack};
+    delete $self->{Parent};
+    delete $self->{DOM};
+
+    $self->{LAST_DOM} = $dom;
+
+    return $dom;
+}
+
+
 sub start_document {
     my ($self, $doc) = @_;
 
@@ -38,12 +55,8 @@ sub xml_decl {
 
 sub end_document {
     my ($self, $doc) = @_;
-    my $dom = $self->{DOM};
-    $dom = $self->{Parent} unless defined $dom; # this is for parsing document chunks
-    delete $self->{NamespaceStack};
-    delete $self->{Parent};
-    delete $self->{DOM};
-    return $dom;
+    my $d = $self->done();
+    return $d;
 }
 
 sub start_prefix_mapping {
@@ -65,7 +78,6 @@ sub start_prefix_mapping {
 sub end_prefix_mapping {
     my $self = shift;
     my $ns = shift;
-
     $self->{NamespaceStack}->undeclare_prefix( $ns->{Prefix} );
 }
 
@@ -80,49 +92,53 @@ sub start_element {
         $self->{NamespaceStack}->push_context;
     }
 
-    if ($el->{NamespaceURI}) {
-        if ( defined $self->{DOM} ) {
-            $node = $self->{DOM}->createElementNS($el->{NamespaceURI},
-                                                  $el->{Name});
-        }
-        else {
-            $node = XML::LibXML::Element->new( $el->{Name} );
-            $node->setNamespace( $el->{NamespaceURI},
-                                 $el->{Prefix} , 1 );
-        }
+    if ( defined $self->{Parent} ) {
+        $el->{NamespaceURI} ||= "";
+        $node = $self->{Parent}->addNewChild( $el->{NamespaceURI},
+                                              $el->{Name} );
     }
     else {
-        if ( defined $self->{DOM} ) {
-            $node = $self->{DOM}->createElement($el->{Name});
+        if ($el->{NamespaceURI}) {
+            if ( defined $self->{DOM} ) {
+                $node = $self->{DOM}->createRawElementNS($el->{NamespaceURI},
+                                                         $el->{Name});
+            }
+            else {
+                $node = XML::LibXML::Element->new( $el->{Name} );
+                $node->setNamespace( $el->{NamespaceURI},
+                                     $el->{Prefix} , 1 );
+            }
         }
         else {
-            $node = XML::LibXML::Element->new( $el->{Name} );
+            if ( defined $self->{DOM} ) {
+                $node = $self->{DOM}->createRawElement($el->{Name});
+            }
+            else {
+                $node = XML::LibXML::Element->new( $el->{Name} );
+            }
         }
+
+        $self->{DOM}->setDocumentElement($node);
     }
 
     # build namespaces
+    my $skip_ns= 0;
     foreach my $p ( $self->{NamespaceStack}->get_declared_prefixes() ) {
+        $skip_ns= 1;
         my $uri = $self->{NamespaceStack}->get_uri($p);
         my $nodeflag = 0;
         if ( defined $uri
              and defined $el->{NamespaceURI}
              and $uri eq $el->{NamespaceURI} ) {
-            $nodeflag = 1;
+            #            $nodeflag = 1;
+            next;
         }
-        $node->setNamespace($uri, $p, $nodeflag );
+        $node->setNamespace($uri, $p, 0 );
     }
 
-    # append
-    if ($self->{Parent}) {
-        $self->{Parent}->appendChild($node);
-        $self->{Parent} = $node;
-    }
-    else {
-        $self->{DOM}->setDocumentElement($node);
-        $self->{Parent} = $node;
-    }
+    $self->{Parent} = $node;
 
-     $self->{NamespaceStack}->push_context;
+    $self->{NamespaceStack}->push_context;
 
     # do attributes
     foreach my $key (keys %{$el->{Attributes}}) {
@@ -137,7 +153,7 @@ sub start_element {
 
 
             if ( defined $attr->{Prefix}
-                 and $attr->{Prefix} eq "xmlns" ) {
+                 and $attr->{Prefix} eq "xmlns" and $skip_ns == 0 ) {
                 # ok, the generator does not set namespaces correctly!
                 my $uri = $attr->{Value};
                 $node->setNamespace($uri,
@@ -163,6 +179,16 @@ sub end_element {
     $self->{Parent} = $self->{Parent}->parentNode();
 }
 
+sub start_cdata {
+    my $self = shift;
+    $self->{IN_CDATA} = 1;
+}
+
+sub end_cdata {
+    my $self = shift;
+    $self->{IN_CDATA} = 0;
+}
+
 sub characters {
     my ($self, $chars) = @_;
     if ( not defined $self->{DOM} and not defined $self->{Parent} ) {
@@ -172,14 +198,28 @@ sub characters {
     }
     return unless $self->{Parent};
     my $node;
+
+    unless ( defined $chars and defined $chars->{Data} ) {
+        return;
+    }
+
     if ( defined $self->{DOM} ) {
-        $node = $self->{DOM}->createTextNode($chars->{Data});
+        if ( defined $self->{IN_CDATA} and $self->{IN_CDATA} == 1 ) {
+            $node = $self->{DOM}->createCDATASection($chars->{Data});
+        }
+        else {
+            $node = $self->{Parent}->appendText($chars->{Data});
+            return;
+        }
+    }
+    elsif ( defined $self->{IN_CDATA} and $self->{IN_CDATA} == 1 ) {
+        $node = XML::LibXML::CDATASection->new($chars->{Data});
     }
     else {
         $node = XML::LibXML::Text->new($chars->{Data});
     }
 
-    $self->{Parent}->appendChild($node);
+    $self->{Parent}->addChild($node);
 }
 
 sub comment {
@@ -191,6 +231,10 @@ sub comment {
         $self->{NamespaceStack}->push_context;
     }
 
+    unless ( defined $chars and defined $chars->{Data} ) {
+        return;
+    }
+
     if ( defined $self->{DOM} ) {
         $comment = $self->{DOM}->createComment( $chars->{Data} );
     }
@@ -199,10 +243,10 @@ sub comment {
     }
 
     if ( defined $self->{Parent} ) {
-        $self->{Parent}->appendChild($comment);
+        $self->{Parent}->addChild($comment);
     }
     else {
-        $self->{DOM}->appendChild($comment);
+        $self->{DOM}->addChild($comment);
     }
 }
 
@@ -213,10 +257,10 @@ sub processing_instruction {
     $PI = $self->{DOM}->createPI( $pi->{Target}, $pi->{Data} );
 
     if ( defined $self->{Parent} ) {
-        $self->{Parent}->appendChild( $PI );
+        $self->{Parent}->addChild( $PI );
     }
     else {
-        $self->{DOM}->appendChild( $PI );
+        $self->{DOM}->addChild( $PI );
     }
 }
 
