@@ -1,6 +1,6 @@
 /**
  * perl-libxml-sax.c
- * $Id: perl-libxml-sax.c,v 1.18 2002/09/12 22:17:25 phish Exp $
+ * $Id: perl-libxml-sax.c,v 1.25 2003/05/22 23:23:43 phish Exp $
  */
 
 #ifdef __cplusplus
@@ -27,14 +27,14 @@ extern "C" {
 #endif
 
 #define NSDELIM ':'
-#define NSDEFAULTURI "http://www.w3.org/XML/1998/namespace"
-
+/* #define NSDEFAULTURI "http://www.w3.org/XML/1998/namespace" */ 
+#define NSDEFAULTURI "http://www.w3.org/2000/xmlns/"
 typedef struct {
     SV * parser;
     xmlNodePtr ns_stack;
     xmlSAXLocator * locator;
     xmlDocPtr ns_stack_root;
-    SV ** handler;
+    SV * handler;
 } PmmSAXVector;
 
 typedef PmmSAXVector* PmmSAXVectorPtr;
@@ -63,12 +63,16 @@ _C2Sv( const xmlChar *string, const xmlChar *dummy )
     STRLEN len;
 
     if ( string != NULL ) {
-        len = xmlStrlen( string );
-        retval = newSVpvn( (const char *)string, len );
+        xmlChar * str = xmlStrdup( string );
+        len = xmlStrlen( str );
+        retval = NEWSV(0, len+1); 
+        sv_setpvn(retval, (const char *)str, len );
 #ifdef HAVE_UTF8
         SvUTF8_on( retval );
 #endif
+        xmlFree( str );
     }
+
     return retval;
 }
 
@@ -87,6 +91,9 @@ PmmSAXInitialize()
     PERL_HASH(EncodingHash,   "Encoding",      8);
 }
 
+xmlSAXHandlerPtr PSaxGetHandler();
+
+
 void
 PmmSAXInitContext( xmlParserCtxtPtr ctxt, SV * parser )
 {
@@ -96,26 +103,31 @@ PmmSAXInitContext( xmlParserCtxtPtr ctxt, SV * parser )
     dTHX;
 
     vec = (PmmSAXVector*) xmlMalloc( sizeof(PmmSAXVector) );
+
     vec->ns_stack_root = xmlNewDoc(NULL);
     vec->ns_stack      = xmlNewDocNode(vec->ns_stack_root,
                                        NULL,
-                                       "stack",
+                                       (const xmlChar*)"stack",
                                        NULL );
 
     xmlAddChild((xmlNodePtr)vec->ns_stack_root, vec->ns_stack);
 
     vec->locator = NULL;
 
-    SvREFCNT_inc( parser );
-    vec->parser  = parser;
+    vec->parser  = SvREFCNT_inc( parser );
     th = hv_fetch( (HV*)SvRV(parser), "HANDLER", 7, 0 );
     if ( th != NULL && SvTRUE(*th) ) {
-        vec->handler = th  ;
+        vec->handler = SvREFCNT_inc(*th)  ;
     }
     else {
         vec->handler = NULL  ;
     }
 
+    if ( ctxt->sax ) {
+        xmlFree( ctxt->sax );
+    }
+    ctxt->sax = PSaxGetHandler();
+
     ctxt->_private = (void*)vec;
 }
 
@@ -125,11 +137,21 @@ PmmSAXCloseContext( xmlParserCtxtPtr ctxt )
     PmmSAXVector * vec = (PmmSAXVectorPtr) ctxt->_private;
     dTHX;
 
-    vec->handler = NULL;
+    if ( vec->handler != NULL ) {
+        SvREFCNT_dec( vec->handler );
+        vec->handler = NULL;
+    }
+
+    xmlFree( ctxt->sax );
+    ctxt->sax = NULL;
+
     SvREFCNT_dec( vec->parser );
+    vec->parser = NULL;
 
     xmlFreeDoc( vec->ns_stack_root );
+    vec->ns_stack_root;
     xmlFree( vec );
+    ctxt->_private = NULL;
 }
 
 
@@ -146,13 +168,19 @@ PmmGetNsMapping( xmlNodePtr ns_stack, const xmlChar * prefix )
 
 void
 PSaxStartPrefix( PmmSAXVectorPtr sax, const xmlChar * prefix,
-                 const xmlChar * uri, SV ** handler )
+                 const xmlChar * uri, SV * handler )
 {
     dTHX;
-    HV * param = newHV();
+    HV * param;
+    SV * rv;
 
     dSP;
 
+    ENTER;
+    SAVETMPS;
+
+    param = newHV();
+
     hv_store(param, "NamespaceURI", 12,
              _C2Sv(uri, NULL), NsURIHash);
 
@@ -162,33 +190,41 @@ PSaxStartPrefix( PmmSAXVectorPtr sax, const xmlChar * prefix,
     }
     else {
         hv_store(param, "Prefix", 6,
-                 _C2Sv("", NULL), PrefixHash);
+                 _C2Sv((const xmlChar*)"", NULL), PrefixHash);
     }
 
-    
-    ENTER;
-    SAVETMPS;
-
     PUSHMARK(SP) ;
-    XPUSHs(*handler);
-    XPUSHs(newRV_noinc((SV*)param));
+    XPUSHs(handler);
+
+    rv = newRV_noinc((SV*)param);
+
+    XPUSHs(rv);
     PUTBACK;
 
-    perl_call_method( "start_prefix_mapping", 0 );
-
+    perl_call_method( "start_prefix_mapping", G_SCALAR | G_EVAL );
+    sv_2mortal(rv);
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
     FREETMPS ;
     LEAVE ;
 }
 
 void
 PSaxEndPrefix( PmmSAXVectorPtr sax, const xmlChar * prefix,
-               const xmlChar * uri, SV ** handler )
+               const xmlChar * uri, SV * handler )
 {
     dTHX;
-    HV * param = newHV();
+    HV * param;
+    SV * rv;
 
     dSP;
 
+    ENTER;
+    SAVETMPS;
+    param = newHV();
     hv_store(param, "NamespaceURI", 12,
              _C2Sv(uri, NULL), NsURIHash);
 
@@ -198,48 +234,85 @@ PSaxEndPrefix( PmmSAXVectorPtr sax, const xmlChar * prefix,
     }
     else {
         hv_store(param, "Prefix", 6,
-                 _C2Sv("", NULL), PrefixHash);
+                 _C2Sv((const xmlChar *)"", NULL), PrefixHash);
     }
-    
-    ENTER;
-    SAVETMPS;
 
     PUSHMARK(SP) ;
-    XPUSHs(*handler);
-    XPUSHs(newRV_noinc( (SV*)param ));
+    XPUSHs(handler);
+
+
+    rv = newRV_noinc((SV*)param);
+
+    XPUSHs(rv);
     PUTBACK;
 
-    perl_call_method( "end_prefix_mapping", 0 );
-
+    perl_call_method( "end_prefix_mapping", G_SCALAR | G_EVAL );
+    sv_2mortal(rv);
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
+    
     FREETMPS ;
     LEAVE ;
 }
 
 void 
 PmmExtendNsStack( PmmSAXVectorPtr sax , const xmlChar * name) {
-    xmlNodePtr newNS = xmlNewDocNode( sax->ns_stack_root, NULL, name, NULL );
-    xmlAddChild(sax->ns_stack, newNS);
-    sax->ns_stack = newNS;
+    xmlNodePtr newNS = NULL;
+    xmlChar * localname = NULL;
+    xmlChar * prefix = NULL;
+    
+    localname = xmlSplitQName( NULL, name, &prefix );
+    if ( prefix != NULL ) {
+        /* check if we can find a namespace with that prefix... */
+        xmlNsPtr ns = xmlSearchNs( sax->ns_stack->doc, sax->ns_stack, prefix );
+
+        if ( ns != NULL ) {
+            newNS = xmlNewDocNode( sax->ns_stack_root, ns, localname, NULL );
+        }
+        else {
+            newNS = xmlNewDocNode( sax->ns_stack_root, NULL, name, NULL );
+        }
+    }
+    else {
+        newNS = xmlNewDocNode( sax->ns_stack_root, NULL, name, NULL );
+    }
+
+    if ( newNS != NULL ) {
+        xmlAddChild(sax->ns_stack, newNS);
+        sax->ns_stack = newNS;
+    }
+
+    if ( localname != NULL ) {
+        xmlFree( localname ) ;
+    }
+    if ( prefix != NULL ) {
+        xmlFree( prefix );
+    }
 }
 
 void
-PmmNarrowNsStack( PmmSAXVectorPtr sax, SV **handler )
+PmmNarrowNsStack( PmmSAXVectorPtr sax, SV *handler )
 {
     xmlNodePtr parent = sax->ns_stack->parent;
     xmlNsPtr list = sax->ns_stack->nsDef;
+
     while ( list ) {
-        if ( !xmlStrEqual(list->prefix, "xml") ) {
+        if ( !xmlStrEqual(list->prefix, (const xmlChar*)"xml") ) {
             PSaxEndPrefix( sax, list->prefix, list->href, handler );
         }
-        list = list->next;
+        list = list->next;        
     }
     xmlUnlinkNode(sax->ns_stack);
+    xmlFreeNode(sax->ns_stack);
     sax->ns_stack = parent;
 }
 
 void
 PmmAddNamespace( PmmSAXVectorPtr sax, const xmlChar * name,
-                 const xmlChar * href, SV **handler)
+                 const xmlChar * href, SV *handler)
 {
     xmlNsPtr ns         = NULL;
     xmlChar * nodename  = NULL;
@@ -251,67 +324,51 @@ PmmAddNamespace( PmmSAXVectorPtr sax, const xmlChar * name,
         return;
     }
 
-    nodename = xmlStrdup(sax->ns_stack->name);
-    localname = xmlSplitQName( NULL, nodename, &prefix );
-    xmlFree( nodename );
+    ns = xmlNewNs( sax->ns_stack, href, name );         
 
-    if ( name != NULL ) {
-        ns = xmlNewNs( sax->ns_stack, href, name );         
-        PSaxStartPrefix( sax, name, href, handler );
-        
-        if ( sax->ns_stack->ns == NULL
-             && xmlStrEqual( prefix , name ) ) {
-            sax->ns_stack->ns = ns;
-            xmlFree( (xmlChar*) sax->ns_stack->name );
-            sax->ns_stack->name = (const xmlChar*) xmlStrdup( localname );
+    if ( sax->ns_stack->ns == NULL ) {
+        localname = xmlSplitQName( NULL, sax->ns_stack->name, &prefix );
+
+        if ( name != NULL ) {
+            if ( xmlStrEqual( prefix , name ) ) {
+                xmlChar * oname = (xmlChar*)(sax->ns_stack->name);
+                sax->ns_stack->ns = ns;
+                xmlFree( oname );
+                sax->ns_stack->name = (const xmlChar*) xmlStrdup( localname );
+            }
         }
-    }
-    else {
-        ns = xmlNewNs( sax->ns_stack, href, NULL );
-        PSaxStartPrefix( sax, NULL, href, handler );
-        if ( prefix = NULL ) {
+        else if ( prefix == NULL ) {
             sax->ns_stack->ns = ns;
         }
     }
 
-    if ( prefix != NULL ) {
+    if ( prefix ) {
         xmlFree( prefix );
     }
-    if ( localname != NULL ) {
+    if ( localname ) {
         xmlFree( localname );
     }
+
+    PSaxStartPrefix( sax, name, href, handler );
 }
 
 HV *
 PmmGenElementSV( pTHX_ PmmSAXVectorPtr sax, const xmlChar * name )
 {
     HV * retval = newHV();
+    SV * tmp;
+    xmlChar * localname = NULL;
+    xmlChar * prefix    = NULL;
 
     xmlNsPtr ns = NULL;
+
     if ( name != NULL && xmlStrlen( name )  ) {
-        xmlChar *localname = NULL, *prefix = NULL;
         hv_store(retval, "Name", 4,
                  _C2Sv(name, NULL), NameHash);
 
-        if ( sax->ns_stack->ns != NULL ) { 
-            localname = xmlSplitQName(NULL, name, &prefix);
- 
-            ns = PmmGetNsMapping( sax->ns_stack, prefix );
-            if ( ns != NULL ) {
-                if ( sax->ns_stack->ns == NULL ) {
-                    sax->ns_stack->ns = ns;
-                    xmlFree( (xmlChar*)sax->ns_stack->name );
-                    sax->ns_stack->name = (const xmlChar*)xmlStrdup(localname);
-                }
-            }
-            xmlFree(prefix);
-            xmlFree(localname);
+        localname = xmlSplitQName(NULL, name, &prefix);
+        ns = PmmGetNsMapping( sax->ns_stack, prefix );
 
-        }
-        else {
-            ns = sax->ns_stack->ns ;
-        }
-        
         if ( ns != NULL ) {
             hv_store(retval, "NamespaceURI", 12,
                      _C2Sv(ns->href, NULL), NsURIHash);
@@ -321,22 +378,22 @@ PmmGenElementSV( pTHX_ PmmSAXVectorPtr sax, const xmlChar * name )
             }
             else {
                 hv_store(retval, "Prefix", 6,
-                         _C2Sv("", NULL), PrefixHash);                
+                         _C2Sv((const xmlChar *)"",NULL), PrefixHash);
             }
 
             hv_store(retval, "LocalName", 9,
                      _C2Sv(sax->ns_stack->name, NULL), LocalNameHash);
-        } 
+        }
         else {
             hv_store(retval, "NamespaceURI", 12,
-                     _C2Sv("", NULL), NsURIHash);
+                     _C2Sv((const xmlChar *)"",NULL), NsURIHash);
             hv_store(retval, "Prefix", 6,
-                     _C2Sv("", NULL), PrefixHash);
+                     _C2Sv((const xmlChar *)"",NULL), PrefixHash);
             hv_store(retval, "LocalName", 9,
                      _C2Sv(name, NULL), LocalNameHash);
-
         }
-    }            
+    }
+
     return retval;
 }
 
@@ -346,24 +403,25 @@ PmmGenNsName( const xmlChar * name, const xmlChar * nsURI )
     int namelen = 0;
     int urilen = 0;
     xmlChar * retval = NULL;
+
+    if ( name == NULL ) {
+        return NULL;
+    }
     namelen = xmlStrlen( name );
+
+    retval =xmlStrncat( retval, (const xmlChar *)"{", 1 );
     if ( nsURI != NULL ) {
         urilen = xmlStrlen( nsURI );
-    }
-
-
-    retval =xmlStrncat( retval, "{",1 );
-    if ( nsURI != NULL ) {
         retval =xmlStrncat( retval, nsURI, urilen );
     } 
-    retval = xmlStrncat( retval, "}",1 );
+    retval = xmlStrncat( retval, (const xmlChar *)"}", 1 );
     retval = xmlStrncat( retval, name, namelen );
     return retval;
 }
 
 HV *
 PmmGenAttributeHashSV( pTHX_ PmmSAXVectorPtr sax,
-                       const xmlChar **attr, SV ** handler )
+                       const xmlChar **attr, SV * handler )
 {
     HV * retval     = NULL;
     HV * atV        = NULL;
@@ -381,8 +439,9 @@ PmmGenAttributeHashSV( pTHX_ PmmSAXVectorPtr sax,
     xmlChar * localname   = NULL;
     xmlChar * prefix      = NULL;
 
+    retval = newHV();
+
     if ( ta != NULL ) {
-        retval = newHV();
         while ( *ta != NULL ) {
             atV = newHV();
             name = *ta;  ta++;
@@ -393,89 +452,91 @@ PmmGenAttributeHashSV( pTHX_ PmmSAXVectorPtr sax,
 
                 hv_store(atV, "Name", 4,
                          _C2Sv(name, NULL), NameHash);
-                hv_store(atV, "Value", 5,
-                         _C2Sv(value, NULL), ValueHash);
-
-                if ( prefix != NULL ) {
-                    if ( xmlStrEqual( "xmlns", prefix ) ) {
-                        PmmAddNamespace( sax,
-                                         localname,
-                                         value,
-                                         handler);
-
-                        nsURI = "http://www.w3.org/2000/xmlns/";
-
-                        hv_store(atV, "Prefix", 6,
-                                 _C2Sv(prefix, NULL), PrefixHash);
-                        hv_store(atV, "LocalName", 9,
-                                 _C2Sv(localname,NULL), LocalNameHash);
-                        hv_store(atV, "NamespaceURI", 12,
-                                 _C2Sv("http://www.w3.org/2000/xmlns/",NULL),
-                                 NsURIHash);
-                    }
-                    else {
-                        ns = PmmGetNsMapping( sax->ns_stack, prefix );
-
-                        if ( ns != NULL ) {
-                            nsURI = ns->href;
-
-                            hv_store(atV, "NamespaceURI", 12,
-                                     _C2Sv(ns->href, NULL), NsURIHash);
-                            hv_store(atV, "Prefix", 6,
-                                     _C2Sv(prefix, NULL), PrefixHash);
-                            hv_store(atV, "LocalName", 9,
-                                     _C2Sv(localname, NULL), LocalNameHash);
-                        }
-                        else {
-                            hv_store(atV, "NamespaceURI", 12,
-                                     _C2Sv("",NULL), NsURIHash);
-                            hv_store(atV, "Prefix", 6,
-                                     _C2Sv("", NULL), PrefixHash);
-                            hv_store(atV, "LocalName", 9,
-                                     _C2Sv(name, NULL), LocalNameHash);
-                        }
-                    }
-                    xmlFree( prefix );
+                if ( value != NULL ) {
+                    hv_store(atV, "Value", 5,
+                             _C2Sv(value, NULL), ValueHash);
                 }
-                else if ( xmlStrEqual( "xmlns", name ) ) {
+
+                if ( xmlStrEqual( (const xmlChar *)"xmlns", name ) ) {
                     /* a default namespace */
                     PmmAddNamespace( sax, NULL, value, handler);  
-                    nsURI = "http://www.w3.org/2000/xmlns/";
+                    /* nsURI = (const xmlChar*)NSDEFAULTURI; */
+                    nsURI = NULL;
+                    hv_store(atV, "Name", 4,
+                             _C2Sv(name, NULL), NameHash);
 
                     hv_store(atV, "Prefix", 6,
-                             _C2Sv(localname, NULL), PrefixHash);
+                             _C2Sv("", NULL), PrefixHash);
                     hv_store(atV, "LocalName", 9,
-                             _C2Sv("",NULL), LocalNameHash);
+                             _C2Sv(name,NULL), LocalNameHash);
                     hv_store(atV, "NamespaceURI", 12,
-                             _C2Sv("http://www.w3.org/2000/xmlns/",NULL),
-                             NsURIHash);
+                             _C2Sv("", NULL), NsURIHash);
                     
                 }
-                else {
+                else if (xmlStrncmp((const xmlChar *)"xmlns:", name, 6 ) == 0 ) {
+                    PmmAddNamespace( sax,
+                                     localname,
+                                     value,
+                                     handler);
+
+                    nsURI = (const xmlChar*)NSDEFAULTURI;
+         
                     hv_store(atV, "Prefix", 6,
-                             _C2Sv("",NULL), PrefixHash);
+                             _C2Sv(prefix, NULL), PrefixHash);
                     hv_store(atV, "LocalName", 9,
                              _C2Sv(localname, NULL), LocalNameHash);
                     hv_store(atV, "NamespaceURI", 12,
-                             _C2Sv("",NULL), NsURIHash);
-                    
+                             _C2Sv((const xmlChar *)NSDEFAULTURI,NULL),
+                             NsURIHash);
                 }
-                
-                keyname = PmmGenNsName( localname, nsURI );
+                else if ( prefix != NULL
+                          && (ns = PmmGetNsMapping( sax->ns_stack, prefix ) ) ) {
+                    nsURI = ns->href;
+
+                    hv_store(atV, "NamespaceURI", 12,
+                             _C2Sv(ns->href, NULL), NsURIHash);
+                    hv_store(atV, "Prefix", 6,
+                             _C2Sv(ns->prefix, NULL), PrefixHash);
+                    hv_store(atV, "LocalName", 9,
+                             _C2Sv(localname, NULL), LocalNameHash);
+                }
+                else {
+                    nsURI = NULL;
+                    hv_store(atV, "NamespaceURI", 12,
+                             _C2Sv((const xmlChar *)"", NULL), NsURIHash);
+                    hv_store(atV, "Prefix", 6,
+                             _C2Sv((const xmlChar *)"", NULL), PrefixHash);
+                    hv_store(atV, "LocalName", 9,
+                             _C2Sv(name, NULL), LocalNameHash);
+                }
+
+                keyname = PmmGenNsName( localname != NULL ? localname : name,
+                                        nsURI );
 
                 len = xmlStrlen( keyname );
-                PERL_HASH( atnameHash, keyname, len );
+                PERL_HASH( atnameHash, (const char *)keyname, len );
                 hv_store(retval,
-                         keyname,
+                         (const char *)keyname,
                          len,
                          newRV_noinc((SV*)atV),
                          atnameHash );
 
-                xmlFree( keyname );
-                xmlFree(localname);
+                if ( keyname != NULL ) {
+                    xmlFree( keyname );
+                }
+                if ( localname != NULL ) {
+                    xmlFree(localname);
+                }
+                localname = NULL;
+                if ( prefix != NULL ) {
+                    xmlFree( prefix );
+                }
+                prefix    = NULL;
+
             }            
         }
     }
+
     return retval;
 }
 
@@ -509,7 +570,7 @@ PmmGenPISV( pTHX_ PmmSAXVectorPtr sax,
         }
         else {
             hv_store(retval, "Data", 4,
-                     _C2Sv("", NULL), DataHash);
+                     _C2Sv((const xmlChar *)"", NULL), DataHash);
         }
     }
 
@@ -523,49 +584,64 @@ PSaxStartDocument(void * ctx)
     PmmSAXVectorPtr sax   = (PmmSAXVectorPtr)ctxt->_private;
     int count             = 0;
     dTHX;
-    HV* real_obj          = (HV *)SvRV(sax->parser);
-    HV* empty             = newHV();
-    SV ** handler         = sax->handler;
+    HV* empty;
+    SV * handler         = sax->handler;
 
+    SV * rv;
     if ( handler != NULL ) {
+       
         dSP;
         
         ENTER;
         SAVETMPS;
         
+        empty = newHV();
         PUSHMARK(SP) ;
-        XPUSHs(*handler);
-        XPUSHs(newRV_noinc((SV*)empty));
+        XPUSHs(handler);
+        XPUSHs(sv_2mortal(newRV_noinc((SV*)empty)));
         PUTBACK;
         
-        count = perl_call_method( "start_document", 0 );
+        count = perl_call_method( "start_document", G_SCALAR | G_EVAL );
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
         
         SPAGAIN;
-        
+
         PUSHMARK(SP) ;
 
     
-        XPUSHs(*handler);
-
+        XPUSHs(handler);
 
+        empty = newHV();
         if ( ctxt->version != NULL ) {
             hv_store(empty, "Version", 7,
                      _C2Sv(ctxt->version, NULL), VersionHash);
         }
         else {
             hv_store(empty, "Version", 7,
-                     _C2Sv("1.0", NULL), VersionHash);
+                     _C2Sv((const xmlChar *)"1.0", NULL), VersionHash);
         }
         
-        if ( ctxt->encoding != NULL ) {
+        if ( ctxt->input->encoding != NULL ) {
             hv_store(empty, "Encoding", 8,
-                     _C2Sv(ctxt->encoding, NULL), EncodingHash);
+                     _C2Sv(ctxt->input->encoding, NULL), EncodingHash);
         }
-        XPUSHs(newRV_noinc((SV*)empty ));
+
+        rv = newRV_noinc((SV*)empty);
+        XPUSHs( rv);
 
         PUTBACK;
         
-        count = perl_call_method( "xml_decl", 0 );
+        count = perl_call_method( "xml_decl", G_SCALAR | G_EVAL );
+        sv_2mortal(rv);
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
         
         FREETMPS ;
         LEAVE ;
@@ -591,8 +667,13 @@ PSaxEndDocument(void * ctx)
     XPUSHs(sax->parser);
     PUTBACK;
 
-    count = perl_call_pv( "XML::LibXML::_SAXParser::end_document", 0 );
-
+    count = perl_call_pv( "XML::LibXML::_SAXParser::end_document", G_SCALAR | G_EVAL );
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
+    
     FREETMPS ;
     LEAVE ;
 
@@ -606,11 +687,12 @@ PSaxStartElement(void *ctx, const xmlChar * name, const xmlChar** attr)
     PmmSAXVectorPtr  sax  = (PmmSAXVectorPtr)ctxt->_private;
     int count             = 0;
     dTHX;
-    SV * attrhash         = NULL;
-    HV * real_obj         = (HV *)SvRV(sax->parser);
+    HV * attrhash         = NULL;
     HV * element          = NULL;
-    SV ** handler         = sax->handler;
-    
+    SV * handler          = sax->handler;
+    SV * rv;
+    SV * arv;
+
     dSP;
     
     ENTER;
@@ -618,36 +700,36 @@ PSaxStartElement(void *ctx, const xmlChar * name, const xmlChar** attr)
 
     PmmExtendNsStack(sax, name);
 
-    attrhash = (SV*) PmmGenAttributeHashSV(aTHX_ sax, attr, handler );
+    attrhash = PmmGenAttributeHashSV(aTHX_ sax, attr, handler );
     element  = PmmGenElementSV(aTHX_ sax, name);
 
-    if ( attrhash != NULL ) {
-        hv_store( element,
-                  "Attributes",
-                  10,
-                  newRV_noinc(attrhash),
-                  AttributesHash );
-    }
-    else {
-        hv_store( element,
-                  "Attributes",
-                  10,
-                  newRV_noinc((SV*)newHV()),
-                  AttributesHash );
-    }
+    arv = newRV_noinc((SV*)attrhash);
+    hv_store( element,
+              "Attributes",
+              10,
+              arv,
+              AttributesHash );
     
     PUSHMARK(SP) ;
-    XPUSHs(*handler);
-    XPUSHs(newRV_noinc((SV*)element));
+
+    XPUSHs(handler);
+    rv = newRV_noinc((SV*)element);
+    XPUSHs(rv);
     PUTBACK;
 
-    count = perl_call_method( "start_element", 0 );
-
-    SPAGAIN ;
+    count = perl_call_method( "start_element", G_SCALAR | G_EVAL );
+    sv_2mortal(rv) ;
 
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
+    
     FREETMPS ;
     LEAVE ;
 
+
     return 1;
 }
 
@@ -655,10 +737,11 @@ int
 PSaxEndElement(void *ctx, const xmlChar * name) {
     xmlParserCtxtPtr ctxt = (xmlParserCtxtPtr)ctx;
     PmmSAXVectorPtr  sax  = (PmmSAXVectorPtr)ctxt->_private;
-    int count             = 0;
     dTHX;
-    HV* real_obj          = (HV *)SvRV(sax->parser);
-    SV ** handler         = sax->handler;
+    int count;
+    SV * handler         = sax->handler;
+    SV * rv;
+    HV * element;
 
     dSP;
 
@@ -666,12 +749,23 @@ PSaxEndElement(void *ctx, const xmlChar * name) {
     SAVETMPS;
 
     PUSHMARK(SP) ;
-    XPUSHs(*handler);
-    XPUSHs(newRV_noinc((SV*)PmmGenElementSV(aTHX_ sax, name)));
+    XPUSHs(handler);
+
+    element = PmmGenElementSV(aTHX_ sax, name);
+    rv = newRV_noinc((SV*)element);
+
+    XPUSHs(rv);
     PUTBACK;
 
-    count = perl_call_method( "end_element", 0 );
-
+    count = perl_call_method( "end_element", G_SCALAR | G_EVAL );
+    sv_2mortal(rv);
+    
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
+    
     FREETMPS ;
     LEAVE ;
 
@@ -686,9 +780,16 @@ PSaxCharacters(void *ctx, const xmlChar * ch, int len) {
     PmmSAXVectorPtr sax = (PmmSAXVectorPtr)ctxt->_private;
     int count = 0;
     dTHX;
-    HV* real_obj = (HV *)SvRV(sax->parser);
-    HV* empty = newHV();
-    SV ** handler = sax->handler;
+    HV* element;
+    SV * handler;
+    SV * rv = NULL;
+
+    if ( sax == NULL ) {
+/*         warn( "lost my sax context!? ( %s, %d )\n", ch, len ); */
+        return 0;
+    }
+
+    handler = sax->handler;
 
     if ( ch != NULL && handler != NULL ) {
         xmlChar * data = xmlStrndup( ch, len );
@@ -699,12 +800,22 @@ PSaxCharacters(void *ctx, const xmlChar * ch, int len) {
         SAVETMPS;
 
         PUSHMARK(SP) ;
-        XPUSHs(*handler);
-        XPUSHs(newRV_noinc((SV*)PmmGenCharDataSV(aTHX_ sax,data)));
+
+        XPUSHs(handler);
+        element = PmmGenCharDataSV(aTHX_ sax,data);
+        rv = newRV_noinc((SV*)element);
+        XPUSHs(rv);
         PUTBACK;
 
-        count = perl_call_method( "characters", 0 );
-
+        count = perl_call_method( "characters", G_SCALAR | G_EVAL );
+        sv_2mortal(rv);
+        
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
         FREETMPS ;
         LEAVE ;
 
@@ -720,9 +831,10 @@ PSaxComment(void *ctx, const xmlChar * ch) {
     PmmSAXVectorPtr sax = (PmmSAXVectorPtr)ctxt->_private;
     int count = 0;
     dTHX;
-    HV* real_obj = (HV *)SvRV(sax->parser);
-    HV* empty = newHV();
-    SV ** handler = sax->handler;
+    HV* element;
+    SV * handler = sax->handler;
+    
+    SV * rv = NULL;
 
     if ( ch != NULL && handler != NULL ) {
         xmlChar * data = xmlStrdup( ch );
@@ -733,12 +845,21 @@ PSaxComment(void *ctx, const xmlChar * ch) {
         SAVETMPS;
 
         PUSHMARK(SP) ;
-        XPUSHs(*handler);
-        XPUSHs(newRV_noinc((SV*)PmmGenCharDataSV(aTHX_ sax,data)));
+        XPUSHs(handler);
+        element = PmmGenCharDataSV(aTHX_ sax,data);
+        rv = newRV_noinc((SV*)element);
+        XPUSHs(rv);
         PUTBACK;
 
-        count = perl_call_method( "comment", 0 );
-
+        count = perl_call_method( "comment", G_SCALAR | G_EVAL );
+        sv_2mortal(rv);
+        
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
         FREETMPS ;
         LEAVE ;
 
@@ -753,11 +874,12 @@ PSaxCDATABlock(void *ctx, const xmlChar * ch, int len) {
     xmlParserCtxtPtr ctxt = (xmlParserCtxtPtr)ctx;
     PmmSAXVectorPtr sax = (PmmSAXVectorPtr)ctxt->_private;
     int count = 0;
-
     dTHX;
-    HV* real_obj = (HV *)SvRV(sax->parser);
-    HV* empty = newHV();
-    SV ** handler = sax->handler;
+
+    HV* element;
+    SV * handler = sax->handler;
+    
+    SV * rv = NULL;
 
     if ( ch != NULL && handler != NULL ) {
         xmlChar * data = xmlStrndup( ch, len );
@@ -768,12 +890,46 @@ PSaxCDATABlock(void *ctx, const xmlChar * ch, int len) {
         SAVETMPS;
 
         PUSHMARK(SP) ;
-        XPUSHs(sax->parser);
-        XPUSHs(newRV_noinc((SV*)PmmGenCharDataSV(aTHX_ sax,data)));
+        XPUSHs(handler);
+        PUTBACK;
+        count = perl_call_method( "start_cdata", G_SCALAR | G_EVAL );
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
+        SPAGAIN;        
+        PUSHMARK(SP) ;
+    
+        XPUSHs(handler);
+        element = PmmGenCharDataSV(aTHX_ sax,data);
+        rv = newRV_noinc((SV*)element);
+        XPUSHs(rv);
         PUTBACK;
 
-        count = perl_call_pv( "XML::LibXML::_SAXParser::cdata_block", 0 );
+        count = perl_call_method( "characters", G_SCALAR | G_EVAL );
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
+        SPAGAIN;        
+        PUSHMARK(SP) ;
+    
+        XPUSHs(handler);
+        PUTBACK;
 
+        count = perl_call_method( "end_cdata", G_SCALAR | G_EVAL );
+        sv_2mortal(rv);
+        
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
         FREETMPS ;
         LEAVE ;
 
@@ -781,6 +937,7 @@ PSaxCDATABlock(void *ctx, const xmlChar * ch, int len) {
     }
 
     return 1;
+
 }
 
 int
@@ -790,9 +947,11 @@ PSaxProcessingInstruction( void * ctx, const xmlChar * target, const xmlChar * d
     PmmSAXVectorPtr sax   = (PmmSAXVectorPtr)ctxt->_private;
     int count             = 0;
     dTHX;
-    HV* real_obj          = (HV *)SvRV(sax->parser);
     HV* empty             = newHV();
-    SV ** handler         = sax->handler;
+    SV * handler          = sax->handler;
+
+    SV * element;
+    SV * rv = NULL;
 
     if ( handler != NULL ) {
         dSP;
@@ -801,13 +960,23 @@ PSaxProcessingInstruction( void * ctx, const xmlChar * target, const xmlChar * d
         SAVETMPS;
 
         PUSHMARK(SP) ;
-        XPUSHs(*handler);
-        XPUSHs(newRV_noinc((SV*)PmmGenPISV(aTHX_ sax, target, data)));
+        XPUSHs(handler);
+        element = (SV*)PmmGenPISV(aTHX_ sax, (const xmlChar *)target, data);
+        rv = newRV_noinc((SV*)element);
+        XPUSHs(rv);
 
         PUTBACK;
 
-        count = perl_call_method( "processing_instruction", 0 );
+        count = perl_call_method( "processing_instruction", G_SCALAR | G_EVAL );
 
+        sv_2mortal(rv);
+
+        if (SvTRUE(ERRSV)) {
+            STRLEN n_a;
+            POPs;
+            croak(SvPV(ERRSV, n_a));
+        }
+        
         FREETMPS ;
         LEAVE ;
     }
@@ -828,7 +997,13 @@ PmmSaxWarning(void * ctx, const char * msg, ...)
     svMessage = NEWSV(0,512);
 
     va_start(args, msg);
-    sv_vsetpvfn(svMessage, msg, xmlStrlen(msg), &args, NULL, 0, NULL);
+    sv_vsetpvfn(svMessage,
+                msg,
+                xmlStrlen((const xmlChar *)msg),
+                &args,
+                NULL,
+                0,
+                NULL);
     va_end(args);
 
     ENTER;
@@ -843,11 +1018,16 @@ PmmSaxWarning(void * ctx, const char * msg, ...)
 
     PUTBACK;
 
-    perl_call_pv( "XML::LibXML::_SAXParser::warning", 0 );
+    perl_call_pv( "XML::LibXML::_SAXParser::warning", G_SCALAR | G_EVAL );
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
     
     FREETMPS ;
     LEAVE ;
-    SvREFCNT_dec(svMessage);
+
     return 1;
 }
 
@@ -864,28 +1044,32 @@ PmmSaxError(void * ctx, const char * msg, ...)
     dTHX;
     dSP;
 
-    svMessage = NEWSV(0,512);
-
-    va_start(args, msg);
-    sv_vsetpvfn(svMessage, msg, xmlStrlen(msg), &args, NULL, 0, NULL);
-    va_end(args);
-
-
     ENTER;
     SAVETMPS;
 
     PUSHMARK(SP) ;
+
     XPUSHs(sax->parser);
 
+    svMessage = NEWSV(0,512);
+
+    va_start(args, msg);
+    sv_vsetpvfn(svMessage, msg, xmlStrlen((const xmlChar *)msg), &args, NULL, 0, NULL);
+    va_end(args);
+
     XPUSHs(sv_2mortal(svMessage));
     XPUSHs(sv_2mortal(newSViv(ctxt->input->line)));
     XPUSHs(sv_2mortal(newSViv(ctxt->input->col)));
     PUTBACK;
-    perl_call_pv( "XML::LibXML::_SAXParser::error", 0 );
+    perl_call_pv( "XML::LibXML::_SAXParser::error", G_SCALAR | G_EVAL );
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
     
     FREETMPS ;
     LEAVE ;
-    SvREFCNT_dec(svMessage);
     return 1;
 }
 
@@ -905,7 +1089,7 @@ PmmSaxFatalError(void * ctx, const char * msg, ...)
     svMessage = NEWSV(0,512);
 
     va_start(args, msg);
-    sv_vsetpvfn(svMessage, msg, xmlStrlen(msg), &args, NULL, 0, NULL);
+    sv_vsetpvfn(svMessage, msg, xmlStrlen((const xmlChar *)msg), &args, NULL, 0, NULL);
     va_end(args);
 
     ENTER;
@@ -918,11 +1102,15 @@ PmmSaxFatalError(void * ctx, const char * msg, ...)
     XPUSHs(sv_2mortal(newSViv(ctxt->input->line)));
     XPUSHs(sv_2mortal(newSViv(ctxt->input->col)));
     PUTBACK;
-    perl_call_pv( "XML::LibXML::_SAXParser::fatal_error", 0 );
+    perl_call_pv( "XML::LibXML::_SAXParser::fatal_error", G_SCALAR | G_EVAL );
+    if (SvTRUE(ERRSV)) {
+        STRLEN n_a;
+        POPs;
+        croak(SvPV(ERRSV, n_a));
+    }
     
     FREETMPS ;
     LEAVE ;
-    SvREFCNT_dec(svMessage);
     return 1;
 }
 
@@ -937,6 +1125,10 @@ PSaxGetHandler()
     memset(retval, 0, sizeof(xmlSAXHandler));
 
     retval->startDocument = (startDocumentSAXFunc)&PSaxStartDocument;
+
+    /* libxml2 will not handle perls returnvalue correctly, so we have 
+     * to end the document ourselfes
+     */
     retval->endDocument   = NULL; /* (endDocumentSAXFunc)&PSaxEndDocument; */
 
     retval->startElement  = (startElementSAXFunc)&PSaxStartElement;
