@@ -3,6 +3,8 @@
 #include <libxml/xpathInternals.h>
 #include <libxml/uri.h>
 
+#include "EXTERN.h"
+
 #include "dom.h"
 
 void
@@ -29,6 +31,11 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         obj2 = valuePop(ctxt);
     }
 
+
+    /* first assure the XML::LibXML error handler is deactivated 
+       otherwise strange things might happen
+     */
+
     if (ctxt->value->type == XPATH_NODESET) {
         int i;
         xmlXPathObjectPtr newobj, ret;
@@ -59,6 +66,9 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         if (obj2 != NULL)
             xmlXPathFreeObject(obj2);
         valuePush(ctxt, ret);
+
+        /* reset the error old error handler before leaving
+         */
         return;
     }
     /*
@@ -69,6 +79,10 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         ctxt->error = XPATH_INVALID_TYPE;
         if (obj2 != NULL)
             xmlXPathFreeObject(obj2);
+
+        /* reset the error old error handler before leaving
+         */
+
         return;
     }
     obj = valuePop(ctxt);
@@ -113,6 +127,9 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
     xmlXPathFreeObject(obj);
     if (obj2 != NULL)
         xmlXPathFreeObject(obj2);
+
+    /* reset the error old error handler before leaving
+     */
 }
 
 
@@ -127,19 +144,45 @@ xmlXPathObjectPtr
 domXPathFind( xmlNodePtr refNode, xmlChar * path ) {
     xmlXPathObjectPtr res = NULL;
   
-    if ( refNode != NULL && refNode->doc != NULL && path != NULL ) {
-        /* we can only do a path in a valid document! 
-         */
+    if ( refNode != NULL && path != NULL ) {
         xmlXPathContextPtr ctxt;
         xmlXPathCompExprPtr comp;
-    
+
+        xmlDocPtr tdoc = NULL;
+        xmlNodePtr froot = refNode;
+
+        comp = xmlXPathCompile( path );
+        if ( comp == NULL ) {
+            return NULL;
+        }
+        
+        if ( refNode->doc == NULL ) {
+            /* if one XPaths a node from a fragment, libxml2 will
+               refuse the lookup. this is not very usefull for XML
+               scripters. thus we need to create a temporary document
+               to make libxml2 do it's job correctly.
+             */
+            tdoc = xmlNewDoc( NULL );
+
+            /* find refnode's root node */
+            while ( froot != NULL ) {
+                if ( froot->parent == NULL ) {
+                    break;
+                }
+                froot = froot->parent;
+            }
+            xmlAddChild((xmlNodePtr)tdoc, froot);
+
+            refNode->doc = tdoc;
+        }
+
         /* prepare the xpath context */
         ctxt = xmlXPathNewContext( refNode->doc );
         ctxt->node = refNode;
-    
         /* get the namespace information */
         if (refNode->type == XML_DOCUMENT_NODE) {
-            ctxt->namespaces = xmlGetNsList(refNode->doc, refNode->children);
+            ctxt->namespaces = xmlGetNsList( refNode->doc,
+                                             xmlDocGetRootElement( refNode->doc ) );
         }
         else {
             ctxt->namespaces = xmlGetNsList(refNode->doc, refNode);
@@ -154,18 +197,27 @@ domXPathFind( xmlNodePtr refNode, xmlChar * path ) {
                              (const xmlChar *) "document",
                              perlDocumentFunction);
        
+        res = xmlXPathCompiledEval(comp, ctxt);
 
-        comp = xmlXPathCompile( path );
-        if (comp != NULL) {
-            res = xmlXPathCompiledEval(comp, ctxt);
-            xmlXPathFreeCompExpr(comp);
-        }
+        xmlXPathFreeCompExpr(comp);
 
         if (ctxt->namespaces != NULL) {
             xmlFree( ctxt->namespaces );
         }
 
         xmlXPathFreeContext(ctxt);
+
+        if ( tdoc != NULL ) {
+            /* after looking through a fragment, we need to drop the
+               fake document again */
+            froot->doc = NULL;
+            tdoc->children = NULL;
+            tdoc->last     = NULL;
+            xmlSetTreeDoc(froot, NULL);
+            refNode->doc = NULL;
+
+            xmlFreeDoc( tdoc );
+        }
     }
     return res;
 }
